#include "ScalarFieldPolygonizer.h"
#include "CellField.h"
#include "CellMesh.h"
#include <interpol/interpol.h>
#include <malloc.h>
#include <math/math.h>

typedef struct {
	uint8_t	rotated_verts[8];
	uint8_t	rotated_edges[12];
} CellAttributesRotated;

static CellAttributesRotated const cellRotatedX = {
	.rotated_verts = {
		0x4, 0x5, 0x1, 0x0, 0x7, 0x6, 0x2, 0x3
	},

	.rotated_edges = {
		0x4, 0x9, 0x0, 0x8, 0x6, 0xA, 0x2, 0xB, 0x7, 0x5, 0x1, 0x3
	}
};

static CellAttributesRotated const cellRotatedY = {
	.rotated_verts = {
		0x1, 0x5, 0x6, 0x2, 0x0, 0x4, 0x7, 0x3
	},

	.rotated_edges = {
		0x9, 0x5, 0xA, 0x1, 0x8, 0x7, 0xB, 0x3, 0x0, 0x4, 0x6, 0x2
	}
};

static CellAttributesRotated const cellRotatedZ = {
	.rotated_verts = {
		0x1, 0x2, 0x3, 0x0, 0x5, 0x6, 0x7, 0x4
	},

	.rotated_edges = {
		0x1, 0x2, 0x3, 0x0, 0x5, 0x6, 0x7, 0x4, 0x9, 0xA, 0xB, 0x8
	}
};

typedef struct {
	uint8_t numTriples;
	struct {
		uint8_t	ids[3];
	} triples[4];
} EdgeIndexTriples;

#define V0	(0x01 << 0)
#define V1	(0x01 << 1)
#define V2	(0x01 << 2)
#define V3	(0x01 << 3)
#define V4	(0x01 << 4)
#define V5	(0x01 << 5)
#define V6	(0x01 << 6)
#define V7	(0x01 << 7)
static EdgeIndexTriples vertexMaskToEdgeIndexTriples [256] = {
	{ 0x0 },
	[V0] = { 0x1, { {{ 0x0, 0x3, 0x8 }} } },	/* point */

	[V6 | V0] = { 0x2, { {{ 0x0, 0x3, 0x8 }}, {{ 0x5, 0xA, 0x6 }} } },	/* cube diagonal */
	[V2 | V0] = { 0x2, { {{ 0x0, 0x3, 0x8 }}, {{ 0x1, 0xA, 0x2 }} } },	/* face diagonal */
	[V3 | V0] = { 0x2, { {{ 0x0, 0x2, 0x8 }}, {{ 0x8, 0x2, 0xB }} } },	/* 1 edge */

	[V3 | V1 | V0] = { 0x3, { {{ 0x9, 0x1, 0x8 }}, {{ 0x1, 0x2, 0x8 }}, {{ 0x2, 0xB, 0x8 }} } },	/* 2 connected edges */
	[V6 | V1 | V0] = { 0x3, { {{ 0x9, 0x3, 0x8 }}, {{ 0x9, 0x1, 0x3 }}, {{ 0xA, 0x5, 0x6 }} } },	/* edge and cube diagonal */
	[V5 | V2 | V0] = { 0x3, { {{ 0x0, 0x3, 0x8 }}, {{ 0x1, 0xA, 0x2 }}, {{ 0x9, 0x1, 0x5 }} } },	/* 2 face diagonals */

	[V3 | V2 | V1 | V0] = { 0x2, { {{ 0x9, 0xB, 0x8 }}, {{ 0x9, 0xA, 0xB }} } },	/* 1 face */
	[V7 | V6 | V1 | V0] = { 0x4, { {{ 0x9, 0x1, 0x8 }}, {{ 0x1, 0x3, 0x8 }}, {{ 0x7, 0xA, 0x5 }}, {{ 0x7, 0xB, 0xA }} } },	/* 2 distant edges */
	[V6 | V3 | V1 | V0] = { 0x4, { {{ 0x9, 0x1, 0x8 }}, {{ 0x1, 0x2, 0x8 }}, {{ 0x2, 0xB, 0x8 }}, {{ 0xA, 0x6, 0x5 }} } },	/* 2 connected edges and cube diagonal */
	[V6 | V3 | V2 | V1] = { 0x4, { {{ 0xB, 0x5, 0x6 }}, {{ 0xB, 0x9, 0x5 }}, {{ 0xB, 0x0, 0x9 }}, {{ 0xB, 0x3, 0x0 }} } },	/* 3 edges with shared point */
	[V7 | V5 | V2 | V0] = { 0x4, { {{ 0x0, 0x3, 0x8 }}, {{ 0x9, 0x4, 0x5 }}, {{ 0x1, 0xA, 0x2 }}, {{ 0x6, 0xB, 0x7 }} } },	/* 2 face diagonals */
	[V7 | V3 | V1 | V0] = { 0x4, { {{ 0x8, 0x6, 0x7 }}, {{ 0x8, 0x2, 0x6 }}, {{ 0x8, 0x1, 0x2 }}, {{ 0x8, 0x9, 0x1 }} } },	/* 3 connected edges */
	[V5 | V3 | V1 | V0] = { 0x4, { {{ 0x8, 0x2, 0xB }}, {{ 0x8, 0x4, 0x2 }}, {{ 0x2, 0x4, 0x1 }}, {{ 0x4, 0x5, 0x1 }} } },	/* 3 connected edges */

	[0xFF & ~(V3 | V1 | V0)] = { 0x3, { {{ 0x9, 0x8, 0x1 }}, {{ 0x1, 0x8, 0x2 }}, {{ 0x2, 0x8, 0xB }} } },	/* 2 connected edges */
	[0xFF & ~(V6 | V1 | V0)] = { 0x3, { {{ 0x9, 0x8, 0x3 }}, {{ 0x9, 0x3, 0x1 }}, {{ 0xA, 0x6, 0x5 }} } },	/* edge and cube diagonal */
	[0xFF & ~(V5 | V2 | V0)] = { 0x3, { {{ 0x0, 0x8, 0x3 }}, {{ 0x1, 0x2, 0xA }}, {{ 0x9, 0x5, 0x1 }} } },	/* 2 face diagonals */

	[0xFF & ~(V6 | V0)] = { 0x2, { {{ 0x0, 0x8, 0x3 }}, {{ 0x5, 0x6, 0xA }} } },	/* cube diagonal */
	[0xFF & ~(V2 | V0)] = { 0x2, { {{ 0x0, 0x8, 0x3 }}, {{ 0x1, 0x2, 0xA }} } },	/* face diagonal */
	[0xFF & ~(V3 | V0)] = { 0x2, { {{ 0x0, 0x8, 0x2 }}, {{ 0x8, 0xB, 0x2 }} } },	/* 1 edge */

	[0xFF & ~V0] = { 0x1, { {{ 0x0, 0x8, 0x3 }} } },	/* point */
};

static void PolygonizeCell( ScalarFieldPolygonizer const* self, Cell const* cell, float treshold, CellMesh* cellMesh ) {
	if( self ) {}

	uint8_t vertexMask = 0x0;
	ScalarFieldVertex* const * vertex = cell->verts;
	for( uint8_t i=0; i<8; ++i ) {
		if( (*vertex)->potential >= treshold ) {
			vertexMask |= (1 << i);
		}
		++vertex;
	}

	EdgeIndexTriples const* triples = &vertexMaskToEdgeIndexTriples[ vertexMask ];
	cellMesh->numTriangles = triples->numTriples;
	CellTriangle* triangle = cellMesh->triangles;
	for( uint8_t t=0; t<triples->numTriples; ++t, ++triangle ) {
		Vertex_pos3f_n3f* vertex = triangle->verts;
		uint8_t const* tripleIndex = triples->triples[t].ids;
		for( uint8_t v=0; v<3; ++v, ++vertex, ++tripleIndex ) {
			CellEdge edge = getEdge_Cell( cell, *tripleIndex );
			vertex->position =
				calculate_LERP3f
				(
					&edge.verts[0]->position,
					&edge.verts[1]->position,
					(treshold - edge.verts[0]->potential) / (edge.verts[1]->potential - edge.verts[0]->potential)
				);
		}

		Vector3f u, v;
		sub_Vector3f( &u, &triangle->verts[1].position, &triangle->verts[0].position );
		sub_Vector3f( &v, &triangle->verts[2].position, &triangle->verts[0].position );
		Vector3f normal;
		crossProduct_Vector3f( &normal, &u, &v );
		normalize_Vector3f( &normal, &normal );

		vertex = triangle->verts;
		for( uint8_t v=0; v<3; ++v, ++vertex ) {
			vertex->normal = normal;
		}
	}
}

static Mesh ConvertCellMeshesToMesh( ScalarFieldPolygonizer const* self, CellMesh const cellMeshes[], size_t cellMeshesLength ) {
	if( self ) {}

	Mesh mesh;
	init_Mesh( &mesh );

	Layer layer;
	init_Layer( &layer );

	CellMesh const* cellMesh = &cellMeshes[0];
	size_t vertexIdx = 0;
	for( size_t k=0; k<cellMeshesLength; ++k, ++cellMesh ) {
		CellTriangle const* cellTriangle = cellMesh->triangles;
		for( size_t j=0; j<cellMesh->numTriangles; ++j, ++cellTriangle ) {
			Vertex_pos3f_n3f const* cellTriangleVertex = cellTriangle->verts;
			for( size_t i=0; i<3; ++i, ++cellTriangleVertex ) {
				Vertex_generic vertex =
				{
					.pos = cellTriangleVertex->position,
					.normal = cellTriangleVertex->normal
				};
				addVertex_Layer( &layer, &vertex );
			}

			Triangle triangle = { .ids = { vertexIdx, vertexIdx+1, vertexIdx+2 } };
			addTriangle_Layer( &layer, &triangle );
			vertexIdx += 3;
		}
	}
	addLayer_Mesh( &mesh, &layer );
	// don't destroy layer as it's owned by mesh

	// don't destroy mesh as it's owned by the caller
	return mesh;
}

static uint8_t RotateCell( uint8_t vertexMask, CellAttributesRotated const* rotated ) {
	uint8_t newVertexMask = 0;
	for( uint8_t i=0; i<8; ++i ) {
		if( vertexMask & (1 << i) ) {
			newVertexMask |= (1 << rotated->rotated_verts[ log2uint( vertexMask & (1 << i) ) ] );
		}
	}

	for( uint8_t tripleIdx=0; tripleIdx<vertexMaskToEdgeIndexTriples[ vertexMask ].numTriples; ++tripleIdx ) {
		vertexMaskToEdgeIndexTriples[ newVertexMask ].numTriples =
			vertexMaskToEdgeIndexTriples[ vertexMask ].numTriples;
		for( uint8_t i=0; i<3; ++i ) {
			vertexMaskToEdgeIndexTriples[ newVertexMask ].triples[ tripleIdx ].ids[ i ]
			 = rotated->rotated_edges[ vertexMaskToEdgeIndexTriples[ vertexMask ].triples[ tripleIdx ].ids[ i ] ];
		}
	}

	return newVertexMask;
}

void init_ScalarFieldPolygonizer( ScalarFieldPolygonizer* self ) {
	self->i = 0;

	for( uint8_t vertexMask=1; vertexMask<255; ++vertexMask ) {
		uint8_t newVertexMaskX, newVertexMaskY, newVertexMaskZ;
		for( uint8_t k=0; k<4; ++k ) {
			if( 0 == k ) { newVertexMaskZ = vertexMask; }
			else { newVertexMaskZ = RotateCell( newVertexMaskZ, &cellRotatedZ ); }

			for( uint8_t j=0; j<4; ++j ) {
				if( 0 == j ) { newVertexMaskY = newVertexMaskZ; }
				else { newVertexMaskY = RotateCell( newVertexMaskY, &cellRotatedY ); }

				for( uint8_t i=0; i<4; ++i ) {
					if( 0 == i ) { newVertexMaskX = newVertexMaskY; }
					newVertexMaskX = RotateCell( newVertexMaskX, &cellRotatedX );
				}
			}
		}
	}
}

Mesh process_ScalarFieldPolygonizer(
	ScalarFieldPolygonizer const* self,
	ScalarField const* scalarField,
	float treshold
) {
	CellField cellField;
	init_CellField( &cellField, scalarField );

	uint16_t n = getNumberOfCells_CellField( &cellField );
	CellMesh* cellMeshes = malloc( sizeof( *cellMeshes ) * n );
	CellMesh* cellMesh = &cellMeshes[0];

	CellFieldIterator iterator = createIterator_CellField( &cellField );
	do {
		Cell const* cell = dereference_CellFieldIterator( &iterator );
		PolygonizeCell( self, cell, treshold, cellMesh );
		++cellMesh;
	} while( next_CellFieldIterator( &iterator ) );
	destroy_CellFieldIterator( &iterator );

	destroy_CellField( &cellField );

	Mesh mesh = ConvertCellMeshesToMesh( self, cellMeshes, n );

	free( cellMeshes );

	return mesh;
}
