#ifndef ENGINE_H
#define ENGINE_H

#include <containers/tree.h>
#include <containers/vec_.h>
#include <dltrans.h>
#include <math/matrix.h>
#include <math/vector.h>
#include <mempool.h>
#include <mesh.h>
#include <myheader.h>
#include <node/billboard.h>
#include <node/camera.h>
#include <node/flare.h>
#include <node/lux.h>
#include <node/xformer.h>
#include <sys/dlist.h>
#include <sys/nativedl.h>
#include <sys/surface.h>
#include <viewport.h>

typedef enum {
	OBJECT_RAW,
	OBJECT_SCENE,
	OBJECT_STATICMESH,
	OBJECT_MESH,
	OBJECT_CAMERA,
	OBJECT_BILLBOARD,
	OBJECT_LUX,
	OBJECT_FLARE
} ObjectType;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
} Raw_Xformer;

typedef struct {
	ObjectType		tag;
	Xformer*		xf;
	StaticMesh*		static_mesh;
} StaticMesh_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Mesh*		mesh;
} Mesh_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Billboard*	billboard;
} Billboard_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Camera*		camera;
} Camera_Xformer;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Lux*		lux;
} Lux_Xformer;

typedef struct {
	StaticMesh_Xformer	static_mesh_xf;
	float				scale;
} Pair_StaticMesh_Xformer_Scale;

typedef struct {
	ObjectType	tag;
	Xformer*	xf;
	Vec*		meshes;	// &Vec(Pair_StaticMesh_Xformer_Scale)
	//StaticMesh_Xformer static_mesh_xf;
} Flare_Xformer;

typedef struct {
	Vec						lights;				// Vec(Lux const*)
	Vec						materials;			// Vec(Material const*)
	MemPool					xformers;
	MemPool					static_meshes;
	MemPool					meshes;
	MemPool					billboards;
	MemPool					cameras;
	MemPool					luxes;
	Vec						lightsXf;			// Vec(Lux_Xformer*)
	Vec						transparentXf;		// Vec(Raw_Xformer*)
	Vec 					opaqueXf;			// Vec(Raw_Xformer*)
	Vec						flaresXf;			// Vec(Flare_Xformer*)
	Vec						shadowCastersXf;	// Vec(Mesh_Xformer*)
	Vec						shadowReceiversXf;	// Vec(Mesh_Xformer*)

	Viewport 				viewport;
	Camera_Xformer const*	camera_xf;
	Camera_Xformer			camera_2d_xf;
	Surface const*			renderTarget;
	Surface const*			zBuffer;
	Vec						cpuMatrixes;
	Matrix					projMtxF;
	DList					dl;
	DLTrans					dltrans;
} Engine;

#include "task.h"

void init_Engine( Engine* self );
void setViewport_Engine( Engine* self, Viewport const* viewport );
void setCamera_Engine( Engine* self, Camera_Xformer const* camera_xf );
void setRenderTarget_Engine( Engine* self, Surface const* renderTarget );
void setZBuffer_Engine( Engine* self, Surface const* zbuffer );
Viewport const* getViewport_Engine( Engine const* self );
Camera_Xformer const* getCamera_Engine( Engine const* self );
Surface const* getRenderTarget_Engine( Engine const* self );
Surface const* getZBuffer_Engine( Engine const* self );
DList* getDisplayList_Engine( Engine* self );
void addLight_Engine( Engine* self, Lux const* light );
void addMaterial_Engine( Engine* self, Material const* material );
void removeLight_Engine( Engine* self, Lux const* light );
void processTask_Engine( Engine* self, Task const* task );

void sceneTaskF_Engine( Engine* self, Tree const* scene );
void flareTaskF_Engine( Engine* self, Tree const* scene );
void overlayTaskF_Engine( Engine* self, void* arg );
void clearColorbufF_Engine( Engine* self, void* arg );
void clearZbufF_Engine( Engine* self, void* arg );

Xformer* createXformer_Engine( Engine* self, ProcessFunc_Xformer processor );
void freeXformer_Engine( Engine* self, Xformer* xf );
Raw_Xformer createRaw_Engine( Engine* self );
StaticMesh_Xformer createStaticMesh_Engine( Engine* self, StaticMesh const* staticMesh );
Mesh_Xformer createMesh_Engine( Engine* self );
void freeMesh_Engine( Engine* self, Mesh_Xformer* mesh_xf );
Billboard_Xformer createBillboard_Engine( Engine* self, StaticMesh const* staticMesh );
Camera_Xformer createCamera_Engine( Engine* self, CameraType type );
Lux_Xformer createLux_Engine( Engine* self );
Flare_Xformer createFlare_Engine( Engine* self, Vec* flare_meshes );

#endif	/* ENGINE_H */

