#ifndef DLIST_H
#define DLIST_H

#include <math/matrix.h>
#include <math/vector.h>
#include <node/lux.h>
#include <material.h>
#include <vertex.h>
#include <viewport.h>
#include <texture.h>
#include <rectangle.h>
#include <types.h>
#include <layer.h>
#include <triangle.h>

typedef uintmax_t	name_t;	// i wonder if this always enforces proper alignment

typedef struct {
	void*	data;
	void*	current;
	size_t	size;
} DList;

#if 0
typedef struct {
	Vector3f	position;
	Vector2f	texcoords;
} Vertex;
#endif
typedef enum {
	PrimTriangle
} DListPrimitiveType;

typedef enum {
	ColorBuffer = 0x01,
	ZBuffer = 0x02,
} DListBufferType;

typedef enum {
	MATRIX_PROJECTION,
	MATRIX_MODELVIEW,
} MatrixType;

typedef enum {
	CM_SHADE,
	CM_DECAL,
} CombineMode;

typedef enum {
	CMA_OPAQUE,
	CMA_TRANSLUCENT
} CombineModeAlpha;

typedef enum {
	RM_ZB_OPAQUE,
	RM_OPAQUE,
	RM_ZB_TRANSLUCENT,
	RM_ZB_TEX_EDGE
} RenderMode;

typedef enum {
	GM_LIGHTING = G_LIGHTING,
	GM_TEXGEN = G_TEXTURE_GEN,
	GM_ZBUFFER = G_ZBUFFER,
	GM_CULLBACK = G_CULL_BACK,
	GM_CULLFRONT = G_CULL_FRONT,
	GM_SHADE = G_SHADE,
	GM_GOURAUD = G_SHADING_SMOOTH,
	GM_FOG = G_FOG,
} GeometryMode;

typedef enum {
	Start,
	SetVertexSource,
	SetTriangleSource,
	LoadVertexBuffer,
	LoadIndexBuffer,
	DrawPrimitives,
	SetBuffer,
	ClearBuffers,
	SetClearColor,
	SetClearZValue,
	LoadTexture,
	Jump,
	End,
	SetTextureScale,
	LoadMatrix,
	LoadLight,
	Sync,
	JumpToTranslated,
	SetCombineMode,
	SetRenderMode,
	NormalizePerspective,
	SetVertexScale,
	SetLookAt,
	SetGeometryMode,
	ClearGeometryMode,
	PipeSync,
	SetViewport,
	SetMaterial,
	SetSegment,
} CommandName;

typedef struct {
	name_t		name;
} Command_Start;

typedef struct {
	name_t		name;
	void const*	data;
	uint8_t		format;
} Command_SetVertexSource;

typedef struct {
	name_t			name;
	Triangle const*	data;
} Command_SetTriangleSource;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nvts;
} Command_LoadVertexBuffer;

typedef struct {
	name_t		name;
	uint32_t	src_ofs;
	uint32_t	dst_ofs;
	uint32_t	nids;
} Command_LoadIndexBuffer;

typedef struct {
	name_t		name;
	uint8_t		type;
	uint32_t	num;
	uint32_t	ofs;
} Command_DrawPrimitives;

typedef struct {
	name_t		name;
	void*		buf;
	uint16_t	width;
	uint16_t	height;
	uint8_t		buftype;
	uint8_t		format;
} Command_SetBuffer;

typedef struct {
	name_t		name;
	uint8_t		types;
} Command_ClearBuffers;

typedef struct {
	name_t		name;
	float		r, g, b, a;
} Command_SetClearColor;

typedef struct {
	name_t		name;
	float		val;
} Command_SetClearZValue;

typedef struct {
	name_t			name;
	Texture const*	texture;
	Rectangle		tile;
	bool			prevent_seams;
} Command_LoadTexture;

typedef struct {
	name_t			name;
	DList const*	dl;
} Command_Jump;

typedef struct {
	name_t		name;
} Command_End;

typedef struct {
	name_t		name;
	float		s;
	float		t;
} Command_SetTextureScale;

typedef struct {
	name_t		name;
	Matrix*		mtx;		// Matrix const* would be better, but libultra gets a pointer to a mutable value. Not sure how unsafe dropping const is.
	MatrixType	type;
} Command_LoadMatrix;

typedef struct {
	name_t			name;
	Lux const*		light;
	Vector3f const*	pos;
	uint16_t		idx;
} Command_LoadLight;

typedef struct {
	name_t		name;
} Command_Sync;

typedef struct {
	name_t		name;
} Command_PipeSync;

typedef struct {
	name_t		name;
	void const*	dl;
} Command_JumpToTranslated;

typedef struct {
	name_t		name;
	CombineMode	mode;
	CombineModeAlpha amode;
} Command_SetCombineMode;

typedef struct {
	name_t		name;
	RenderMode	mode;
} Command_SetRenderMode;

typedef struct {
	name_t		name;
	float		scale;
} Command_NormalizePerspective;

typedef struct {
	name_t		name;
	float		scale;
} Command_SetVertexScale;

typedef struct {
	name_t		name;
	Vector3f	eye;
	Vector3f	object;
	Vector3f	up;
} Command_SetLookAt;

typedef struct {
	name_t			name;
	GeometryMode	gmode;
} Command_SetGeometryMode;

typedef struct {
	name_t			name;
	GeometryMode	gmode;
} Command_ClearGeometryMode;

typedef struct {
	name_t			name;
	Viewport const*	vp;
} Command_SetViewport;

typedef struct {
	name_t		name;
	Material	mtl;
} Command_SetMaterial;

typedef struct {
	name_t		name;
	uint8_t		segment;
	void*		adr;
} Command_SetSegment;

void init_DList( DList* self );
void destroy_DList( DList* self );
void reset_DList( DList* self );
void* getData_DList( DList const* dlist );
size_t getSize_DList( DList const* dlist );

void pushStart_DList( DList* dlist );
void pushSetVertexSource_DList( DList* dlist, void const* vts, uint8_t format );
void pushSetTriangleSource_DList( DList* dlist, Triangle const* triangles );
void pushLoadVertexBuffer_DList( DList* dlist, uint32_t src_ofs,
	uint32_t dst_ofs, uint32_t nvts );
void pushLoadIndexBuffer_DList( DList* dlist, uint32_t src_ofs,
	uint32_t dst_ofs, uint32_t nids );
void pushDrawPrimitives_DList( DList* dlist, uint8_t type, uint32_t num, uint32_t ofs );
void pushSetBuffer_DList( DList* dlist, uint8_t buftype, uint8_t format,
	uint16_t width, uint16_t height, void* buf );
void pushClearBuffers_DList( DList* dlist, uint8_t types );
void pushSetClearColor_DList( DList* dlist, float r, float g, float b, float a );
void pushSetClearZValue_DList( DList* dlist, float val );
void pushLoadTexture_DList( DList* dlist, Texture const* texture, Rectangle tile, bool prevent_seams );
void pushJump_DList( DList* dlist, DList const* dl );
void pushEnd_DList( DList* dlist );
void pushSetTextureScale_DList( DList* dlist, float s, float t);
void pushLoadMatrix_DList( DList* dlist, Matrix* mtx, MatrixType type );
void pushLoadLight_DList( DList* dlist, Lux const* light,
	Vector3f const* position, uint16_t idx );
void pushSync_DList( DList* dlist );
void pushPipeSync_DList( DList* dlist );
void pushJumpToTranslated_DList( DList* dlist, void const* dl );
void pushSetCombineMode_DList( DList* dlist, CombineMode mode, CombineModeAlpha amode );
void pushSetRenderMode_DList( DList* dlist, RenderMode mode );
void pushNormalizePerspective_DList( DList* dlist, float scale );
void pushSetVertexScale_DList( DList* dlist, float scale );
void pushSetLookAt_DList( DList* dlist, Vector3f const* eye, Vector3f const* object, Vector3f const* up );
void pushSetGeometryMode_DList( DList* dlist, GeometryMode gmode );
void pushClearGeometryMode_DList( DList* dlist, GeometryMode gmode );
void pushSetViewport_DList( DList* dlist, Viewport const* vp );
void pushSetMaterial_DList( DList* dlist, Material const* material );
void pushSetSegment_DList( DList* dlist, uint8_t segment, void* adr );

#endif	/* DLIST_H */

