#include <abort.h>
#include <assert.h>
#include <malloc.h>
#include <sys/region.h>
#include <containers/tree.h>

inline TreeNode* new_TreeNode( void const* data, void (*destroy_func)(void*, void*) ) {
	TreeNode* node = malloc( sizeof(*node) );
	assert( NULL != node );
	init_TreeNode( node, data, destroy_func );

	return node;
}

inline void delete_TreeNode( TreeNode* self ) {
	destroy_TreeNode( self );
	free( self );
}

inline void init_TreeNode( TreeNode* self, void const* data, void (*destroy_func)(void*, void*) ) {
	init_Vec( &self->children, sizeof(TreeNode*), ALIGN_DEFAULT, destroy_func, NULL );
	self->parent = NULL;
	self->data = data;
}

inline void destroy_TreeNode( TreeNode* self ) {
	destroy_Vec( &self->children );
}

TreeNode clone_TreeNode( TreeNode const* self, void* (*clone_func)(void const*) ) {
	TreeNode cloned;
	init_TreeNode( &cloned,
		clone_func ? clone_func( self->data ) : self->data,
		self->children.destroy_func
	);

	return cloned;
}

TreeNode* cloneRecursive_TreeNode( TreeNode const* self, void* (*clone_func)(void const*) ) {
	TreeNode* cloned = new_TreeNode( self->data, self->children.destroy_func );
	TreeNode const* child;

	VecConstIterator iter = createConstIterator_Vec( &self->children );
	while( (child = dereference_VecConstIterator( &iter )) ) {
		attachChild_TreeNode( cloned,
			cloneRecursive_TreeNode( child, clone_func )
		);

		next_VecConstIterator( &iter );
	}
	destroy_VecConstIterator( &iter );

	return cloned;
}

inline void attachChild_TreeNode( TreeNode* self, TreeNode* child ) {
	pushBack_Vec( &self->children, &child );
	child->parent = self;
}

inline uint_t getNumChildren_TreeNode( TreeNode const* self ) {
	return getLength_Vec( &self->children );
}

inline TreeNode* getParent_TreeNode( TreeNode const* self ) {
	return self->parent;
}

inline TreeNode* getChild_TreeNode( TreeNode const* self, uint_t i ) {
	return *(TreeNode* const *)getConstElement_Vec( &self->children, i );
}

inline void const* getData_TreeNode( TreeNode const* self ) {
	return self->data;
}

inline void init_Tree( Tree* self, TreeNode* root, void* (*clone_func)(void const*), void (*destroy_func)(void*) ) {
	self->root = root;
	self->clone_func = clone_func;
	self->destroy_func = destroy_func;
}

inline void destroy_Tree( Tree* self ) {
	TreeNodeIterator iter;
	init_TreeNodeIterator( &iter, self->root );

	TreeNode* node;
	while( (node = inc_TreeNodeIterator( &iter )) ) {
		if( self->destroy_func ) { self->destroy_func( node ); }
	}

	destroy_TreeNodeIterator( &iter );
}

inline Tree clone_Tree( Tree const* self ) {
	Tree cloned;
	return cloned;
}

inline TreeNode* getRoot_Tree( Tree const* self ) {
	return self->root;
}

inline void init_TreeNodeIterator( TreeNodeIterator* self, TreeNode* node ) {
	init_Stack( &self->stack, sizeof(self->i), NULL );
	reset_TreeNodeIterator( self, node );
}

inline void destroy_TreeNodeIterator( TreeNodeIterator* self ) {
	destroy_Stack( &self->stack );
}

TreeNode* inc_TreeNodeIterator( TreeNodeIterator* self ) {
	TreeNode* node;
i_believe_in_goto:	node = self->next;
	if( NULL == node ) {
		return NULL;
	}

	bool node_visited = self->i;

	uint_t n = getNumChildren_TreeNode( node );
	if( self->i < n ) {
		self->next = getChild_TreeNode( node, self->i );
		self->i++;
		push_Stack( &self->stack, &self->i );

		self->i = 0;
	} else {
		if( true == isEmpty_Stack( &self->stack ) ) {
			if( 0 != n ) {
				return NULL;
			}

			/* it was a single-node tree */
			self->next = NULL;
			return self->node;
		}
	
		self->i = *(uint_t*)pop_Stack( &self->stack );
		self->next = getParent_TreeNode( node );
	}

	if( node_visited ) {
		goto i_believe_in_goto;
	}

	return node;
}

inline void reset_TreeNodeIterator( TreeNodeIterator* self, TreeNode* node ) {
	self->node = node;
	self->next = self->node;
	self->i = 0;
}

