#include <abort.h>
#include <malloc.h>
#include <string.h>
#include <containers/vec_.h>

Vec* new_Vec( size_t el_size, size_t alignment, void (*destroy_func)(void*, void*), void* destroy_func_arg ) {
	Vec* vec = malloc( sizeof(*vec) );
	if(NULL == vec)
		abort();
	init_Vec( vec, el_size, alignment, destroy_func, destroy_func_arg );
	return vec;
}

inline void delete_Vec( Vec* self ) {
	destroy_Vec( self );
	free( self );
}

void init_Vec( Vec* self, size_t el_size, size_t alignment, void (*destroy_func)(void*, void*), void* destroy_func_arg ) {
	self->el_size = el_size;
	self->alignment = alignment;
	self->num_el = 0;
	self->max_el = 16;
	if( NULL == (
		self->data_unaligned =
		 	malloc( el_size*self->max_el + (1<<alignment) )
		)
	) {
		abort();
	}
	self->end = self->data = ALIGN_ADR( self->data_unaligned, alignment );
	self->destroy_func = destroy_func;
	self->destroy_func_arg = destroy_func_arg;
}

void destroy_Vec( Vec* self ) {
	if( self->destroy_func ) {
		void* element;
		VecIterator elements = createIterator_Vec( self );
		while( (element = dereference_VecIterator( &elements )) ) {
			self->destroy_func( element, self->destroy_func_arg );
			next_VecIterator( &elements );
		}

		destroy_VecIterator( &elements );
	}

	free( self->data_unaligned );
}

Vec clone_Vec( Vec const* self, void* (*clone_func)(void const*) ) {
	Vec cloned;
	init_Vec( &cloned, self->el_size, self->alignment, self->destroy_func, self->destroy_func_arg );

	void const* element;
	VecConstIterator iter = createConstIterator_Vec( self );
	while( (element = dereference_VecConstIterator( &iter )) ) {
		pushBack_Vec( &cloned, clone_func( element ) );
		next_VecConstIterator( &iter );
	}

	destroy_VecConstIterator( &iter );

	return cloned;
}

void resize_Vec( Vec* self, uint32_t newsize ) {
	if( NULL == (
		self->data_unaligned =
			realloc( self->data_unaligned, self->el_size*newsize + (1<<self->alignment) )
		)
	) {
		self->data = NULL;
		self->end = NULL;
		abort();
	}
	self->data = ALIGN_ADR( self->data_unaligned, self->alignment );
	self->end = (uint8_t*)self->data + self->el_size * self->num_el;
	self->max_el = newsize;
}

void pushBack_Vec( Vec* self, void const* element ) {
	if( self->num_el >= self->max_el ) {
		resize_Vec( self, self->max_el << 1 );
	}

	memcpy( self->end, element, self->el_size );
	self->end = (uint8_t*)self->end + self->el_size;
	self->num_el++;
}

inline void* getElement_Vec( Vec* self, uint32_t i ) {
	return (uint8_t*)self->data + i * self->el_size;
}

inline void const* getConstElement_Vec( Vec const* self, uint32_t i ) {
	return (uint8_t*)self->data + i * self->el_size;
}

inline void* getHead_Vec( Vec* self ) {
	return self->data;
}

inline void const* getHeadConst_Vec( Vec const* self ) {
	return self->data;
}

inline void* getTail_Vec( Vec* self ) {
	return (uint8_t*)self->data + self->num_el * self->el_size;
}

inline uint32_t getLength_Vec( Vec const* self ) {
	return self->num_el;
}

inline size_t getElementSize_Vec( Vec const* self ) {
	return self->el_size;
}

inline void reset_Vec( Vec* self ) {
	self->num_el = 0;
	self->end = self->data;
}

inline VecIterator createIterator_Vec( Vec* self ) {
	VecIterator it;
	init_VecIterator( &it, self );

	return it;
}

inline void init_VecIterator( VecIterator* self, Vec* vec ) {
	self->vec = vec;
	self->current_idx = 0;
	self->num_elements = getLength_Vec( vec );
	self->current_element = getHead_Vec( vec );
}

inline void destroy_VecIterator( VecIterator* self ) {
	if( self ) {}
}

inline void* dereference_VecIterator( VecIterator const* self ) {
	if( self->current_idx < self->num_elements ) {
		void* element = self->current_element;
		return element;
	} else {
		return NULL;
	}
}

inline bool next_VecIterator( VecIterator* self ) {
	if( self->current_idx < self->num_elements ) {
		self->current_element = (void*)((uintptr_t)self->current_element + getElementSize_Vec( self->vec ));
		self->current_idx++;
		return true;
	} else {
		return false;
	}
}

inline VecConstIterator createConstIterator_Vec( Vec const* self ) {
	VecConstIterator it;
	init_VecConstIterator( &it, self );

	return it;
}

inline void init_VecConstIterator( VecConstIterator* self, Vec const* vec ) {
	self->vec = vec;
	self->current_idx = 0;
	self->num_elements = getLength_Vec( vec );
	self->current_element = getHeadConst_Vec( vec );
}

inline void destroy_VecConstIterator( VecConstIterator* self ) {
	if( self ) {}
}

inline void const* dereference_VecConstIterator( VecConstIterator const* self ) {
	if( self->current_idx < self->num_elements ) {
		void const* element = self->current_element;
		return element;
	} else {
		return NULL;
	}
}

inline bool next_VecConstIterator( VecConstIterator* self ) {
	if( self->current_idx < self->num_elements ) {
		self->current_element = (void const*)((uintptr_t)self->current_element + getElementSize_Vec( self->vec ));
		self->current_idx++;
		return true;
	} else {
		return false;
	}
}
